# -*- coding: utf-8 -*-

# Copyright (c) Pilot Systems, 2013

# This file is part of Faceted SeSQL.

# SeSQL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# SeSQL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Faceted SeSQL.  If not, see <http://www.gnu.org/licenses/>.

from threading import RLock
from data import *
from django.db.models import signals
import time, string
from django.conf import settings
from django.contrib.contenttypes.models import ContentType

class FacetedSortOption(object):
    """
    An option for a search
    """
    def __init__(self, name, field_name, title, order = 'asc'):
        self.name = name
        self.field_name = field_name
        self.title = title
        self.order = order

class FacetedSearchConfiguration(object):
    """
    This class describes a faceted search configuration
    """
    def __init__(self, model, facets, sort_options = [], default_sort = []):
        """
        Constructor
        """
        self.lock = RLock()
        self.model = model
        ct = ContentType.objects.get_for_model(model)
        allowed = set(string.digits + string.ascii_letters + '_')
        self.name = ("%s_%s" % (ct.app_label, ct.name)).encode('ascii', 'replace')
        self.name = ''.join([ c for c in self.name if c in allowed ])
        self.facets = facets
        self.expire_keys = []
        self.facets_dict = dict([ (f.name, f) for f in facets ])
        self.data = None
        self.default_sort = default_sort
        signals.post_save.connect(self.data_changed, model)
        signals.post_delete.connect(self.data_changed, model)
        signals.m2m_changed.connect(self.m2m_changed)
        self.mc_client = getattr(settings, "FACETED_SESQL_MEMCACHE", None)
        if self.mc_client:
            self.mc_key = settings.FACETED_SESQL_MEMCACHE_PREFIX + "_" + self.name
            self.mc_revision = '0'
            self.mc_client.add(self.mc_key, self.mc_revision)

        self.sort_options = []
        for opt in sort_options:
            if isinstance(opt, (str, unicode)):
                if opt[0] == '-':
                    order = 'desc'
                    opt = opt[1:]
                else:
                    order = 'asc'
                name = order + '_' + opt
                facet = self.facets_dict[opt]
                opt = FacetedSortOption(name, facet.field, facet.title, order)
            self.sort_options.append(opt)

    def __getitem__(self, key):
        """
        Get a facet by name
        """
        return self.facets_dict[key]

    def get_expire_keys(self):
        """
        Get all the keys for auto-data expiring based on facets
        """
        return [ facet.get_expire_key() for facet in self.facets ]

    def get_initial_data(self):
        """
        Get the initial data containing everything
        """
        self.lock.acquire()
        try:
            if self.mc_client:
                rev = self.mc_client.get(self.mc_key)
                if rev != self.mc_revision:
                    self.data = None
                    self.mc_revision = rev
            expire_keys = self.get_expire_keys()
            if self.expire_keys != expire_keys:
                self.data = None
            if not self.data:
                self.expire_keys = expire_keys
                self.data = FacetedData(self)
            data = self.data
        finally:
            self.lock.release()
        return data.make_copy()

    def flush_data(self):
        """
        Flush the loaded data
        """
        if not self.mc_client:
            self.lock.acquire()
            self.data = None
            self.lock.release()
        else:
            self.mc_client.incr(self.mc_key)

    def __iter__(self):
        """
        Iterate over the facets
        """
        return iter(self.facets)

    def data_changed(self, **kwargs):
        """
        Hrm, data changed ? Ok, well, flush data cache
        """
        self.flush_data()
        
    def m2m_changed(self, **kwargs):
        """
        Hrm, m2m changed ? Ok, well, flush data cache if it's our m2m
        """
        if kwargs['model'] == self.model:
            self.flush_data()

    
