# -*- coding: utf-8 -*-

# Copyright (c) Pilot Systems, 2013

# This file is part of Faceted SeSQL.

# SeSQL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# SeSQL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Faceted SeSQL.  If not, see <http://www.gnu.org/licenses/>.

from utils import OrderedSet
from sesql.longquery import longquery
from django.db.models import Q
import copy
import search

class FacetedData(object):
    """
    This is the main data store for a faceted search
    """
    def __init__(self, config, source = None):
        """
        Constructor
        """
        self.config = config
        self.search = None
        self.facets = {}
        self.items = {}
        self.matching = OrderedSet()
        self.matching_without = dict()
        self.frozen = False
        self.need_counting = False

        if source:
            # Ok, handle smart copy constructor
            for name, facet in source.facets.items():
                self.facets[name] = copy.copy(facet)
                self.facets[name].data = self
                self.facets[name].values = dict(facet.values)
            self.items = source.items
            self.sorted_by = source.sorted_by
            self.matching = OrderedSet(source.matching)
        else:
            self.recompute()

    def __getitem__(self, facet):
        """
        Get a facet
        """
        return self.facets[facet]

    def recompute(self):
        """
        Compute the data from backend
        """
        self.sorted_by = {}
        self.items = {}
        self.facets = {}
        self.matching = OrderedSet()
        for facet in self.config:
            self.facets[facet.name] = FacetedDataFacet(self, facet)
        fields = set([ facet.facet.get_sql_field() for facet in self.facets.values() ])
        fields = tuple([ f for f in fields if f ])
        query = longquery(Q(classname = self.config.model), fields = fields, order=self.config.default_sort)
        for item in query.brains():
            self.items[item['id']] = item
            self.matching.add(item['id'])
        for option in self.config.sort_options:
            if not option.field_name:
                continue
            allitems = self.items.values()
            allitems.sort(key = lambda item: item[option.field_name])
            if option.order == 'desc':
                allitems.reverse()
            self.sorted_by[option.name] = [ item['id'] for item in allitems ]
        self.recount()

    def make_copy(self):
        """
        Make a copy of this faceted search
        """
        return self.__class__(self.config, source = self)

    def freeze_counting(self):
        """
        Freeze the automatic update of counts, when doing batch filtering
        """
        self.frozen = True

    def unfreeze_counting(self):
        """
        Freeze the automatic update of counts, when doing batch filtering
        """
        self.frozen = False
        if self.need_counting:
            self.recount()

    def filter_by(self, name, criteria):
        """
        Filter objects on this criteria
        """
        self.matching_without[name] = OrderedSet(self.matching)
        facet = self.config[name]
        todel = set()
        for key, item in self.items.items():
            value = facet.get_value(item)
            if not facet.match(value, criteria):
                todel.add(key)
        self.matching -= todel
        for key, value in self.matching_without.items():
            if key != name:
                value -= todel
        self.recount()

    def recount(self):
        """
        Recount the availability of various options
        """
        if self.frozen:
            self.need_counting = True
        else:
            for facet in self.facets.values():
                facet.recount()
            self.need_counting = False

    def brains(self):
        """
        Iterate over matching content - return brains
        """
        for key in self.matching:
            yield self.items[key]

    def all_brains(self):
        """
        Iterate over all content - return brains
        """
        return self.items.itervalues()

    def brains_without(self, facet):
        """
        Iterate over matching content without given facet - return brains
        """
        matching = self.matching_without.get(facet, self.matching)
        for key in matching:
            yield self.items[key]

    def brains_for(self, facet):
        """
        Iterate over matching content for given facet - return brains
        """
        if not self.matching:
            return self.all_brains()
        else:
            return self.brains_without(facet)

    def __iter__(self):
        """
        Iterate over matching content - return real objects
        """
        for key in self.matching:
            yield self.config.model.objects.get(id = key)

    def sorted(self, order):
        """
        Iterated over sorted content
        """
        if not order or not order.name in self.sorted_by:
            return self.__iter__()
        return self.get_sorted_by(order.name)

    def get_sorted_by(self, name):
        """
        Get data ordered by given order name - must exist
        """
        for key in self.sorted_by[name]:
            if key in self.matching:
                yield self.config.model.objects.get(id = key)

    def __len__(self):
        """
        Count matching content
        """
        return len(self.matching)

    def apply_external_filter(self, fltr):
        """
        Applies a filter to the matching items
        A filter is a function that takes the item and returns True or False,
        items will only be kept if the filter says True
        Warning : this can be slow
        """
        keys = OrderedSet(self.matching)
        for val in self.matching_without.values():
            keys |= set(val)
        todel = set()
        for key in keys:
            item = self.config.model.objects.get(id = key)
            if not fltr(item):
                todel.add(key)
        if todel:
            self.matching -= todel
            for val in self.matching_without.values():
                val -= todel
            self.recount()

class FacetValue(object):
    """
    This is a facet value
    """
    def __init__(self, facet, value, **kwargs):
        """
        Constructor
        """
        self.value = value
        self.facet = facet
        for arg, val in kwargs.items():
            setattr(self, arg, val)
        if self.facet.criteria:
            self.checked = self.facet.facet.do_match(value, self.facet.criteria)

class FacetedDataFacet(object):
    """
    This is an actual facet bound to a dataset
    """
    def __init__(self, data, facet):
        """
        Build from configuration
        """
        self.facet = facet
        self.data = data
        self.criteria = None
        self.min_value = None
        self.max_value = None
        self.values = {}

    def recount(self):
        """
        Recount all available values
        """
        if self.data.search is not None:
            self.criteria = self.data.search.get_criteria(self.facet.name)
        else:
            self.criteria = None
        
        self.min_value = None
        self.max_value = None
        self.values = {}
        
        for item in self.data.brains_for(self.facet.name):
            value = self.facet.get_value(item)
            for val in value:
                current = self.values.get(val, 0)
                self.values[val] = current + 1
                if self.min_value is None or self.min_value > val:
                    self.min_value = val
                if self.max_value is None or self.max_value < val:
                    self.max_value = val

    def get_values(self):
        """
        Get the display value and count
        """
        facets = []
        for value, count in self.values.items():
            title = self.facet.get_display_value(value, self.data)
            facet = FacetValue(self,
                             value = value,
                             title = title,
                             count = count)
            facets.append(facet)
        sorted_facets = sorted(facets, key=lambda item: item.title)
        return sorted_facets

    def start(self):
        """
        This is for range widgets, will return the current starting value
        """
        if self.criteria:
            return max(int(self.criteria[0]), self.min_value)
        return self.min_value

    def end(self):
        """
        This is for range widgets, will return the current ending value
        """
        if self.criteria:
            return min(int(self.criteria[1]), self.max_value)
        return self.max_value

    def get_first_criteria(self):
        """
        Get the first criteria
        """
        return self.criteria[0]
