# -*- coding: utf-8 -*-

# Copyright (c) Pilot Systems, 2013-2015

# This file is part of Faceted SeSQL.

# SeSQL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# SeSQL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Faceted SeSQL.  If not, see <http://www.gnu.org/licenses/>.
from django.utils.encoding import smart_str

from datetime import datetime, date, timedelta

from sesql.longquery import longquery
from django.db.models import Q

class Facet(object):
    """
    A generic base class for facets
    """
    def __init__(self, name, title = None, field = None):
        """
        Constructor
        """
        self.name = name
        self.title = title or name
        self.field = field or name

    def get_sql_field(self):
        """
        Get the SQL field for this facet
        """
        return self.name

    def get_expire_key(self):
        """
        Expire key can be used by the facets to signal the criteria changed,
        especially for date-based criteria
        """
        return None
        
    def get_value(self, item):
        """
        Get the value this criteria has on given item (SeSQL "brain")

        Will always return a list, so we handle mono-valued and
        multi-valued fields the same way
        """
        value = item.get(self.field, [])
        if value is None:
            value = []
        if not isinstance(value, (list, tuple)):
            value = [ value ]
        return value
            
    def get_display_value(self, value, context):
        """
        Get the display value for this database value
        """
        return value

    def match(self, value, criteria):
        """
        Check if this value match this criteria
        """
        if criteria is None:
            return True
        for val in value:
            if self.do_match(val, criteria):
                return True
            
    def do_match(self, value, criteria):
        """
        Internal matching function - will be provided single value and cleaned
        criteria
        """
        return value in criteria

    def get_filter_from_request(self, request):
        """
        Extract the filter value from request
        """
        val = request.GET.getlist(self.name, None)
        if val is not None:
            if not isinstance(val, (list, tuple)):
                val = [ val ]
        return val

class StringFacet(Facet):
    """
    A simple facet, directly displayed as stored in the database
    """
    widget = "checkbox"

    def do_match(self, value, criteria):
        """
        Internal matching function - will be provided single value and cleaned
        criteria
        """
        value = smart_str(value)
        for crit in criteria:
            if value == smart_str(crit):
                return True
        return False

class ChoiceFacet(StringFacet):
    """
    A facet that maps the db values to display values
    """
    def __init__(self, name, mapping, title = None, field = None):
        """
        Constructor
        """
        super(ChoiceFacet, self).__init__(name, title, field)
        self.mapping = mapping

    def get_display_value(self, value, context):
        """
        Apply the mapping
        """
        # We use a cache of the mapping on the search context, to avoid
        # unnecessary computing of it
        key = "-%s-mapping" % self.name
        mapping = getattr(context, key, None)
        if mapping is None:
            if callable(self.mapping):
                mapping = self.mapping()
            elif isinstance(self.mapping, (list, tuple)):
                mapping = dict(self.mapping)
            else:
                mapping = self.mapping
            setattr(context, key, mapping)
        return mapping.get(value, value)

class RangeFacet(Facet):
    """
    A facet that allows to input a range of integers/floats
    """
    widget = "range"

    def do_match(self, value, criteria):
        """
        Check if this value match this criteria
        """
        return int(criteria[0]) <= int(value) <= int(criteria[1])
    
    def get_filter_from_request(self, request):
        """
        Extract the filter value from request
        """
        val = super(RangeFacet, self).get_filter_from_request(request)
        bounds = request.GET.getlist(self.name + '_bounds', None)
        if val == bounds:
            # If the filter we submitted encompass all the previous bounds,
            # then assume it's not enabled
            return None
        return val

class ForeignkeyFacet(ChoiceFacet):
    """
    A facet for a foreignkey, input will be id, displayed value
    the __unicode__ of the object, or given field
    """
    def __init__(self, name, model, subfield = None, title = None, field = None):
        """
        Constructor
        """
        self.model = model
        self.subfield = subfield
        super(ForeignkeyFacet, self).__init__(name, self.get_mapping, 
                                              title = title, field = field)

    def get_mapping(self):
        """
        Get the mapping
        """
        if self.subfield:
            vals = self.model.objects.values('id', self.subfield)
            return dict([ (v['id'], v[self.subfield]) for v in vals ])
        else:
            vals = self.model.objects.all()
            return dict([ (v.id, unicode(v)) for v in vals ])


class DateRangeFacet(ChoiceFacet):
    """
    A facet that will display date ranges such as "today", "last week", ...

    The available ranges should be provided, as a callback function
    so it can recompute based on current date

    The callback function should return a list of :
    (id, display value, start, end)
    start or end can be date objects or None (means to be ignored)
    """
    def get_expire_key(self):
        """
        Expire key can be used by the facets to signal the criteria changed,
        especially for date-based criteria
        """
        return date.today().strftime('%Y-%m-%d')
        
    def get_mapping(self):
        """
        Get the mapping (from cache if possible)
        """
        old_key = getattr(self, '_old_key', None)
        if old_key != self.get_expire_key():
            self._old_key = self.get_expire_key()
            self._mapping = self.mapping()
        return self._mapping

    def get_value(self, item):
        """
        Get the value this criteria has on given item (SeSQL "brain")

        Will always return a list, so we handle mono-valued and
        multi-valued fields the same way
        """
        value = item.get(self.field, None)
        if isinstance(value, datetime):
            value = value.date()
        if not value:
            return []

        res = []
        mapping = self.get_mapping()

        for key, label, start, end in mapping:
            if start and value < start:
                continue
            if end and value > end:
                continue
            res.append(key)
        return res

    def get_display_value(self, value, context):
        """
        Apply the mapping
        """
        mapping = self.get_mapping()

        for key, label, start, end in mapping:
            if key == value:
                return label
        return value

class FTSFacet(Facet):
    """
    A simple facet, directly displayed as stored in the database
    """
    widget = "input"

    def get_value(self, item):
        """
        We only get the id of the object, we'll do a sesql search anyway
        """
        return [ (item['classname'], item['id']) ]
            
    def do_match(self, value, criteria):
        """
        Internal matching function - will be provided single value and cleaned
        criteria
        """
        criteria = criteria[0]
        if not criteria.strip():
            return True
            
        opts = { self.field + '__containswords': criteria }
        res = longquery(Q(classname=value[0])& Q(id=value[1]) & Q(**opts))
        return res.count() > 0

    def get_sql_field(self):
        """
        Get the SQL field for this facet
        """
        return None

